import { isArrayLike, asis, noop } from '@proc7ts/primitives';
import { nextCall, isNextCall, NextCall__symbol } from '@proc7ts/call-thru';

/**
 * Checks whether the given iterable is revertible.
 *
 * This is always `true` for arrays.
 *
 * @param iterable  Iterable to check.
 *
 * @returns `true` if `iterable` has a `reverse` property, or `false` otherwise.
 */
function itsRevertible(iterable) {
    return 'reverse' in iterable;
}

/**
 * Starts iteration over the given `iterable`.
 *
 * This is a shortcut for `iterable[Symbol.iterator]` to make it friendlier to minification.
 *
 * @param iterable  An iterable to iterate over.
 *
 * @return An iterator.
 */
function itsIterator(iterable) {
    return iterable[Symbol.iterator]();
}
/**
 * Builds an iterable iterator over the given `iterable`.
 *
 * @param iterable  An iterable to iterate over.
 *
 * @return An iterator.
 */
function itsIterable(iterable) {
    return function* () { yield* iterable; }();
}
function makeIt(iterate, reverse) {
    const iterable = {
        [Symbol.iterator]: iterate,
    };
    if (!reverse) {
        return iterable;
    }
    const reversible = iterable;
    reversible.reverse = reverse;
    return reversible;
}

/**
 * @packageDocumentation
 * @module @proc7ts/a-iterable
 */
/**
 * Constructs a reversed iterable.
 *
 * If the `source` iterable is an array-like structure, then uses `reverseArray()` function to revert the constructed
 * iterable.
 * If the `source` iterable is revertible, then uses its `reverse()` method to revert the constructed one.
 * Otherwise stores elements to array and reverts them with `reverseArray()` function.
 *
 * @param source  A source iterable.
 *
 * @returns An iterable of the `source` elements in reverse order.
 */
function reverseIt(source) {
    if (isArrayLike(source)) {
        return reverseArray(source);
    }
    if (itsRevertible(source)) {
        const reversed = source.reverse();
        return makeIt(() => itsIterator(reversed));
    }
    return reverseArray(Array.from(source));
}
/**
 * Constructs an iterable of array-like structure elements in reverse order.
 *
 * @param array  Source array.
 *
 * @returns An iterable of the `source` elements in reverse order.
 */
function reverseArray(array) {
    return makeIt(function* () {
        const len = array.length;
        for (let i = len - 1; i >= 0; --i) {
            yield array[i];
        }
    });
}

/**
 * @packageDocumentation
 * @module @proc7ts/a-iterable
 */
/**
 * Performs the given `action` for each element of the given `iterable`.
 *
 * @typeparam T  A type of `iterable` elements.
 * @param iterable  An iterable of elements to perform actions on.
 * @param action  An action to perform on each iterable element. This is a function accepting an element as its only
 * parameter.
 */
function itsEach(iterable, action) {
    for (const element of iterable) {
        action(element);
    }
}
/**
 * Checks whether the given `iterable` is empty.
 *
 * @param iterable  Iterable to check for elements.
 *
 * @return `true` if the given iterable contains at least one element, or `false` otherwise.
 */
function itsEmpty(iterable) {
    // noinspection PointlessBooleanExpressionJS
    return !!itsIterator(iterable).next().done;
}
/**
 * Tests whether all elements of the given `iterable` pass the test implemented by the provided function.
 *
 * @typeparam T  A type of `iterable` elements.
 * @param iterable  An iterable to test elements of.
 * @param test  A predicate function to test each element. Returns `true` to continue tests, or `false` to stop it
 * and return `false` from the method call. It accepts the tested element as the only parameter.
 *
 * @returns `true` if the `test` function returned a truthy value for every element, or `false` otherwise.
 * Returns `true` for empty iterable.
 */
function itsEvery(iterable, test) {
    for (const element of iterable) {
        if (!test(element)) {
            return false;
        }
    }
    return true;
}
/**
 * Tests whether at least one element of the given `iterable` passes the test implemented by the provided function.
 *
 * @typeparam T  A type of `iterable` elements.
 * @param iterable  An iterable to test elements of.
 * @param test  A predicate function to test each element. Returns `false` to continue tests, or `true` to stop it
 * and return `true` from the method call. It accepts the tested element as the only parameter.
 *
 * @returns `true` if the callback function returned a truthy value for at least one element in the array, or `false`
 * otherwise. Returns `false` for empty iterable.
 */
function itsSome(iterable, test) {
    for (const element of iterable) {
        if (test(element)) {
            return true;
        }
    }
    return false;
}
/**
 * Returns the first element of the given `iterable`.
 *
 * @typeparam T  A type of `iterable` elements.
 * @param iterable  Iterable to extract element from.
 *
 * @return Either the first element, or `undefined` if the given `iterable` is empty.
 */
function itsFirst(iterable) {
    const result = itsIterator(iterable).next();
    return !result.done ? result.value : undefined;
}
/**
 * Returns the last element of the given `iterable`.
 *
 * If the given `iterable` is an array-like structure, then just returns its last element. If it is revertible,
 * then extracts the first element of reverted one. Otherwise iterates over elements to find the last one.
 *
 * @typeparam T  A type of `iterable` elements.
 * @param iterable  Iterable to extract element from.
 *
 * @return Either the last element, or `undefined` if the given `iterable` is empty.
 */
function itsLast(iterable) {
    if (isArrayLike(iterable)) {
        return iterable[iterable.length - 1];
    }
    if (itsRevertible(iterable)) {
        return itsFirst(iterable.reverse());
    }
    let last;
    for (const element of iterable) {
        last = element;
    }
    return last;
}
/**
 * Applies a function against an accumulator and each element of the given `iterable` to reduce elements to a single
 * value.
 *
 * @typeparam T  A type of `iterable` elements.
 * @typeparam R  A type of reduced value.
 * @param iterable  An iterable to reduce values of.
 * @param reducer  A function to apply the value returned from the previous `reducer` call and to each element.
 * @param initialValue  Initial value passed to the first `reducer` call.
 *
 * @return Reduced value returned from the last `reducer` call, or `initialValue` if there is no elements in the given
 * `iterable`.
 */
function itsReduction(iterable, reducer, initialValue) {
    let reduced = initialValue;
    for (const element of iterable) {
        reduced = reducer(reduced, element);
    }
    return reduced;
}

/**
 * @packageDocumentation
 * @module @proc7ts/a-iterable
 */
/**
 * Creates a next call in {@link IterableCallChain iterable call chain} that performs the next passes for each
 * element of the given iterable.
 *
 * This call passes elements to the next call on demand. While the `nextEach()` one transforms them all at once,
 * and iterates over results after that.
 *
 * @param iterable  An iterable containing elements to pass down the chain.
 *
 * @returns Next call for iterable call chain.
 */
function nextIterate(iterable) {
    return nextCall((chain, pass) => chain.iterate(pass, iterable));
}

/**
 * @packageDocumentation
 * @module @proc7ts/a-iterable
 */
function filterIt(source, test) {
    return makeIt(function* () {
        for (const element of source) {
            if (test(element)) {
                yield element;
            }
        }
    });
}
function flatMapIt(source, convert = asis) {
    return makeIt(function* () {
        for (const element of source) {
            yield* convert(element);
        }
    });
}
/**
 * Creates a new iterable with the results of calling a provided function on every element of the `source` one.
 *
 * @typeparam T  A type of source elements.
 * @typeparam R  A type of converted elements.
 * @param source  A source iterable.
 * @param convert  A function that produces an element of the new iterable, taking the source element as the only
 * parameter.
 */
function mapIt(source, convert) {
    return makeIt(function* () {
        for (const element of source) {
            yield convert(element);
        }
    });
}

/**
 * @packageDocumentation
 * @module @proc7ts/a-iterable
 */
function thruIt(it, ...passes) {
    let result = [];
    const chain = (index) => {
        const lastPass = index >= passes.length;
        ++index;
        const pass = index < passes.length ? passes[index] : noop;
        const handleResult = (callResult, arg) => {
            if (isNextCall(callResult)) {
                callResult[NextCall__symbol](chain(index), pass);
            }
            else if (lastPass) {
                result.push([arg]);
            }
            else {
                chain(index).pass(pass, callResult);
            }
        };
        return ({
            call(fn, args) {
                handleResult(fn(...args), args);
            },
            pass(fn, arg) {
                handleResult(fn(arg), arg);
            },
            skip() { },
            iterate(fn, iterable) {
                result.push({
                    *[Symbol.iterator]() {
                        for (const item of iterable) {
                            const oldResult = result;
                            const newResult = [];
                            try {
                                result = newResult;
                                handleResult(fn(item), item);
                            }
                            finally {
                                result = oldResult;
                            }
                            for (const res of newResult) {
                                yield* res;
                            }
                        }
                    },
                });
            },
        });
    };
    chain(0).iterate(passes[0], it);
    // eslint-disable-next-line @typescript-eslint/no-unsafe-return
    return flatMapIt(result, asis);
}

/**
 * @internal
 */
const ArrayLikeIterableMethods = [
    'every',
    'filter',
    'flatMap',
    'forEach',
    'map',
    'reduce',
    'reverse',
    'some',
];
/**
 * Abstract `Iterable` implementation with array-like iteration operations.
 *
 * @typeparam T  A type of elements.
 */
class AIterable {
    /**
     * Returns an iterable without elements.
     *
     * @returns An empty iterable instance.
     */
    static none() {
        return noneAIterable; // eslint-disable-line @typescript-eslint/no-use-before-define
    }
    /**
     * Checks whether the given iterable is an array-like one.
     *
     * @param source  An iterable to check.
     *
     * @returns `true` is the `source` has all `ArrayLikeIterable` methods (like `Array` or `AIterable` instance),
     * or `false` otherwise.
     */
    static is(source) {
        return ArrayLikeIterableMethods.every(name => name in source);
    }
    static of(source) {
        if (AIterable.is(source)) {
            return source;
        }
        return AIterable.from(source);
    }
    /**
     * Creates an `AIterable` instance that iterates over the same elements as the given one.
     *
     * Uses [[reverseIt]] function to reverse the constructed iterable.
     *
     * @param source  A source iterable.
     *
     * @return Always new `AIterable` instance.
     */
    static from(source) {
        return makeAIterable(() => source, () => reverseIt(source));
    }
    /**
     * Tests whether all elements pass the test implemented by the provided function.
     *
     * Corresponds to `Array.prototype.every()`.
     *
     * @param test  A predicate function to test each element. Returns `true` to continue tests, or `false` to stop it
     * and return `false` from the method call. It accepts the tested element as the only parameter.
     *
     * @returns `true` if the `test` function returned a truthy value for every element, or `false` otherwise.
     * Returns `true` for empty iterable.
     */
    every(test) {
        return itsEvery(this, test);
    }
    filter(test) {
        return makeAIterable(() => filterIt(this, test), () => filterIt(this.reverse(), test));
    }
    /**
     * First maps each element using a mapping function, then flattens the result into a new iterable.
     *
     * Corresponds to `Array.prototype.flatMap()`.
     *
     * Note that the overridden `flatMap` method of `ArrayLikeIterable` expects an array to be returned from `convert`
     * callback, while in this method it may return arbitrary iterable.
     *
     * @typeparam R  A type of converted elements.
     * @param convert  A function that produces a new iterable, taking the source element as the only parameter.
     *
     * @returns A new [[AIterable]] with each element being the flattened result of the `convert` function call.
     */
    flatMap(convert) {
        return makeAIterable(() => flatMapIt(this, convert), () => flatMapIt(this.reverse(), element => reverseIt(convert(element))));
    }
    /**
     * Performs the given `action` for each element.
     *
     * Corresponds to `Array.prototype.forEach()`.
     *
     * @param action  An action to perform on each iterable element. This is a function accepting an element as its only
     * parameter.
     */
    forEach(action) {
        itsEach(this, action);
    }
    /**
     * Creates a new iterable with the results of calling a provided function on every element.
     *
     * Corresponds to `Array.prototype.map()`.
     *
     * @typeparam R  A type of converted elements.
     * @param convert  A function that produces an element of the new iterable, taking the source element as the only
     * parameter.
     *
     * @return A new [[AIterable]] with each element being the result of the `convert` function call.
     */
    map(convert) {
        return makeAIterable(() => mapIt(this, convert), () => mapIt(this.reverse(), convert));
    }
    /**
     * Applies a function against an accumulator and each element to reduce elements to a single value.
     *
     * Corresponds to `Array.prototype.reduce()`.
     *
     * @typeparam R  A type of reduced value.
     * @param reducer  A function to apply the value returned from the previous `reducer` call and to each element.
     * @param initialValue  Initial value passed to the first `reducer` call.
     *
     * @return Reduced value returned from the last `reducer` call, or `initialValue` if there is no elements in this
     * iterable.
     */
    reduce(reducer, initialValue) {
        return itsReduction(this, reducer, initialValue);
    }
    /**
     * Constructs an iterable containing this iterable's elements in reverse order.
     *
     * By default this method converts iterable to array and then reverts its elements with [[reverseArray]] function.
     *
     * @return Reversed [[AIterable]] instance.
     */
    reverse() {
        return makeAIterable(() => reverseArray(Array.from(this)), () => this);
    }
    /**
     * Tests whether some element passed the test implemented by the provided function.
     *
     * Corresponds to `Array.prototype.some()`.
     *
     * @param test  A predicate function to test each element. Returns `false` to continue tests, or `true` to stop it
     * and return `true` from the method call. It accepts the tested element as the only parameter.
     *
     * @returns `true` if the `test` function returned a truthy value for some element, or `false` otherwise.
     * Returns `false` for empty iterable.
     */
    some(test) {
        return itsSome(this, test);
    }
    thru(...fns) {
        const thru = thruIt;
        return makeAIterable(() => thru(this, ...fns));
    }
}
/**
 * @internal
 */
class NoneAIterable extends AIterable {
    *[Symbol.iterator]() { }
    reverse() {
        return this;
    }
}
/**
 * @internal
 */
const noneAIterable = ( /*#__PURE__*/new NoneAIterable());
/**
 * @internal
 */
function makeAIterable(iterate, reverse) {
    class Iterable extends AIterable {
        [Symbol.iterator]() {
            return itsIterator(iterate());
        }
        reverse() {
            if (!reverse) {
                return super.reverse();
            }
            return AIterable.from(makeIt(() => itsIterator(reverse()), () => this));
        }
    }
    return new Iterable();
}
/**
 * Extends an iterable class with `AIterable` API.
 *
 * @typeparam C  A type of iterable class to extend.
 * @typeparam E  A type of elements to iterate.
 * @param iterableClass  A class to extend.
 *
 * @returns A new class extending original `iterableClass` and implementing the missing [[AIterable]] methods.
 */
function toAIterable(iterableClass) {
    class ExtendedIterable extends iterableClass {
    }
    const extended = ExtendedIterable;
    const proto = extended.prototype;
    ArrayLikeIterableMethods.forEach(name => {
        if (!(name in proto)) {
            Object.defineProperty(proto, name, {
                configurable: true,
                value: AIterable.prototype[name],
            });
        }
    });
    return extended;
}

/**
 * @packageDocumentation
 * @module @proc7ts/a-iterable
 */
/**
 * @internal
 */
function* arrayIterator(array) {
    // eslint-disable-next-line @typescript-eslint/prefer-for-of
    for (let i = 0; i < array.length; ++i) {
        yield array[i];
    }
}
/**
 * Builds a {@link RevertibleIterable revertible iterable} over elements of array-like structure.
 *
 * @param array  An array-like structure. E.g. `Array`, DOM `NodeList`, etc.
 *
 * @returns A revertible iterable over array elements.
 */
function overArray(array) {
    return makeIt(() => arrayIterator(array), () => reverseArray(array));
}
/**
 * @internal
 */
const noneIterable = {
    *[Symbol.iterator]() { },
    reverse() { return this; },
};
/**
 * Returns an iterable without elements.
 *
 * @typeparam T  A type of constructed iterable elements.
 *
 * @returns An empty iterable instance revertible to itself.
 */
function overNone() {
    // eslint-disable-next-line @typescript-eslint/no-unsafe-return
    return noneIterable;
}

/**
 * @packageDocumentation
 * @module @proc7ts/a-iterable
 */
/**
 * Builds an iterable over the keys of the given object.
 *
 * A list of keys is constructed using `Reflect.ownKeys()`.
 *
 * @param target  An object to select keys from.
 */
function overKeys(target) {
    return overArray(Reflect.ownKeys(target));
}
/**
 * Builds an iterable over the key/value entries of the given object.
 *
 * A list of keys is constructed using `Reflect.ownKeys()`.
 *
 * @param target  An object to select keys and values from.
 */
function overEntries(target) {
    const keys = overKeys(target);
    function mapToEntries(_keys) {
        return mapIt(_keys, key => [key, target[key]]);
    }
    return makeIt(() => itsIterator(mapToEntries(keys)), () => mapToEntries(keys.reverse()));
}

export { AIterable, filterIt, flatMapIt, itsEach, itsEmpty, itsEvery, itsFirst, itsIterable, itsIterator, itsLast, itsReduction, itsRevertible, itsSome, makeIt, mapIt, nextIterate, overArray, overEntries, overKeys, overNone, reverseArray, reverseIt, thruIt, toAIterable };
//# sourceMappingURL=a-iterable.mjs.map
